#!/usr/bin/perl

use strict;
use Sendmail::Milter;
use Getopt::Long;
use File::Basename;
use Sys::Syslog qw(:DEFAULT setlogsock);

use check_mail_address;

my($version) = '0.01';
my($author) = 'Victor Ustugov aka corvax';
my($email) = 'victor@corvax.kiev.ua';
my($info) = "verify mail address perl mail filter $version 2003 by $author $email";

my($default_syslog_facility) = 'local4';
my($default_rcpt_tempfail_accept) = 0;
my($default_max_attempts) = 1;
my($verbose, $max_attempts, $use_vrfy, $rcpt_tempfail_accept);

my(@sendmail_config_defaults) = (
    '/etc/mail/sendmail.cf',
    '/etc/sendmail.cf'
);

my($script_name) = basename($0);

my(%syslog_param) = (
    'socket' => 'unix',
    'ident' => $script_name,
    'options' => 'cons,pid',
#    'facility' => 'local4',
    'facility' => 'mail',
    'priority' => 'notice',
#  logformat  :
# %mailbox	-  mailbox
# %domain	-  
# %SMTP_host	- primary MX
# %SMTP_command	- ,  MTA  
# %SMTP_answer	-  MTA
# %check_result	-   
#    'logformat' => 'check %mailbox@%domain: %SMTP_command: %SMTP_answer; %check_result'
#    'logformat' => '%check_result (%mailbox@%domain: %SMTP_command => %SMTP_answer)'
    'logformat' => '%check_result (%mailbox@%domain: %SMTP_host => %SMTP_command => %SMTP_answer)'
);

sub mailfrom_callback {
    my $ctx = shift;
    my $mail_from = shift;

    my($href) = $ctx->getpriv();
#    $$href{'mail from'} = $mail_from;
#    $$href{'auth_authen'} = $ctx->getsymval('{auth_authen}');
    $$href{'auth_type'} = $ctx->getsymval('{auth_type}');
    $ctx->setpriv($href);

    print(STDERR "mailfrom_callback: got mail from:$mail_from\n") if ($verbose);
    
    if ($$href{'auth_type'}) {
	print(STDERR "mailfrom_callback: got auth_type: '".$$href{'auth_type'}."'\n") if ($verbose);
	print(STDERR "mailfrom_callback: skip verify sender address for authenticated sender\n") if ($verbose);
    } else {
        my($command, $answer) = verify_address($mail_from, $use_vrfy, $max_attempts, \%syslog_param, $verbose);

	$command =~ s/[\r\n]+$//;
	$answer =~ s/[\r\n]+$//;
        $mail_from =~ s/^\s*\<//;
	$mail_from =~ s/\>\s*$//;

        print(STDERR "mailfrom_callback: returned command: $command\n") if ($verbose);
	print(STDERR "mailfrom_callback: returned answer:  $answer\n") if ($verbose);

        my($error_code, $enh_error_code, $error_msg, $reply);
	if ($answer =~ /^(\d\d\d)\s+(\d\.\d\.\d)\s+(.+)$/) {
	    $error_code = $1;
	    $enh_error_code = $2;
	    $error_msg = $3;
	} elsif ($answer =~ /^(\d\d\d)\s+(.+)$/) {
	    $error_code = $1;
	    $error_msg = $2;
	    $enh_error_code = '';
	}
        $error_msg =~ s/[\r\n]$//;
	if ($error_code =~ /^5\d\d/) {
	    $reply = "Sender verify fail for <$mail_from>: response to \"$command\" was: $answer";
	    print(STDERR "mailfrom_callback: setreply: '550', '5.7.1', '$reply'\n") if ($verbose);
    	    $ctx->setreply('550', '5.7.1', $reply);
	    return SMFIS_REJECT;
	} elsif ($error_code =~ /^4\d\d/) {
	    $reply = "Could not complete sender verify callout for <$mail_from>: $answer";
	    print(STDERR "mailfrom_callback: setreply: '451', '4.4.0', '$reply'\n") if ($verbose);
    	    $ctx->setreply('451', '4.5.1', $reply);
	    return SMFIS_TEMPFAIL;
	} elsif ($error_code !~ /^2\d\d/) {
	    $reply = "Sender verify fail for <$mail_from>: $answer";
	    print(STDERR "mailfrom_callback: setreply: '550', '5.7.1', '$reply'\n") if ($verbose);
    	    $ctx->setreply('550', '5.7.1', $reply);
	    return SMFIS_REJECT;
	}
    }

    return SMFIS_CONTINUE;
}

sub rcptto_callback {
    my $ctx = shift;
    my $rcpt_to = shift;

    my($href) = $ctx->getpriv();
    $ctx->setpriv($href);

    print(STDERR "rcptto_callback: got rcpt to:$rcpt_to\n") if ($verbose);
    
    if ($$href{'auth_type'}) {
	print(STDERR "rcptto_callback: skip verify recipient address for authenticated sender\n") if ($verbose);
    } else {
        my($command, $answer) = verify_address($rcpt_to, $use_vrfy, $max_attempts, \%syslog_param, $verbose);

	$command =~ s/[\r\n]+$//;
	$answer =~ s/[\r\n]+$//;
        $rcpt_to =~ s/^\s*\<//;
	$rcpt_to =~ s/\>\s*$//;

        print(STDERR "rcptto_callback: returned command: $command\n") if ($verbose);
	print(STDERR "rcptto_callback: returned answer:  $answer\n") if ($verbose);

        my($error_code, $enh_error_code, $error_msg, $reply);
	if ($answer =~ /^(\d\d\d)\s+(\d\.\d\.\d)\s+(.+)$/) {
	    $error_code = $1;
	    $enh_error_code = $2;
	    $error_msg = $3;
	} elsif ($answer =~ /^(\d\d\d)\s+(.+)$/) {
	    $error_code = $1;
	    $error_msg = $2;
	    $enh_error_code = '';
	}
        $error_msg =~ s/[\r\n]$//;
	if ($error_code =~ /^5\d\d/) {
#	    $reply = $error_msg;
#	    $reply =~ s/^\s*\<$rcpt_to\>\.*\s*//;

	    $error_code = '554';
	    $enh_error_code = '5.1.1';
	    $reply = 'Unroutable address';

	    print(STDERR "rcptto_callback: setreply: '$error_code', '$enh_error_code', '$reply'\n") if ($verbose);
    	    $ctx->setreply($error_code, $enh_error_code, $reply);
	    return SMFIS_REJECT;
	} elsif ($error_code =~ /^4\d\d/) {
	    unless ($rcpt_tempfail_accept) {
	        $reply = "Could not complete recipient verify callout for <$rcpt_to>: $answer";
	        print(STDERR "rcptto_callback: setreply: '451', '4.4.0', '$reply'\n") if ($verbose);
    	        $ctx->setreply('451', '4.5.1', $reply);
    		return SMFIS_TEMPFAIL;
	    }
	} elsif ($error_code !~ /^2\d\d/) {
#	    $reply = $error_msg;
#	    $reply =~ s/^\s*\<$rcpt_to\>\.*\s*//;

	    $error_code = '554';
	    $enh_error_code = '5.1.1';
	    $reply = 'Unroutable address';

	    print(STDERR "rcptto_callback: setreply: '$error_code', '$enh_error_code', '$reply'\n") if ($verbose);
    	    $ctx->setreply($error_code, $enh_error_code, $reply);
	    return SMFIS_REJECT;
	}
    }

#    $$href{'rcpt to'} .= ", " if ($$href{'rcpt to'});
#    $$href{'rcpt to'} .= $rcpt_to;
#    $ctx->setpriv($href);

    return SMFIS_CONTINUE;
}

sub abort_callback {
    my $ctx = shift;
    $ctx->setpriv(undef);
    return SMFIS_CONTINUE;
}

sub close_callback {
    my $ctx = shift;
    $ctx->setpriv(undef);
    return SMFIS_CONTINUE;
}

my %my_callbacks = (
#	'connect' =>	\&connect_callback,
#    'helo'	=> \&helo_callback,
    'envfrom'	=> \&mailfrom_callback,
    'envrcpt'	=> \&rcptto_callback,
#    'header'	=> \&header_callback,
#    'eoh'	=> \&eoh_callback,
#    'body'	=> \&body_callback,
#    'eom'	=> \&eom_callback,
    'abort'	=> \&abort_callback,
    'close'	=> \&close_callback,
);

BEGIN: {
    my($milter_name, $sendmail_config, $help, $version);

    #    
    GetOptions (
        'help'			=> \$help,
        'version'		=> \$version,
        'milter-name=s'		=> \$milter_name,
        'sendmail-config=s'	=> \$sendmail_config,
        'syslog-facility=s'	=> \$syslog_param{'facility'},
        'verbose'		=> \$verbose,
	'use-vrfy'		=> \$use_vrfy,
        'max-attempts=i'	=> \$max_attempts,
        'rcpt-tempfail=s'	=> \$rcpt_tempfail_accept,
    );

    if ($help) {
        print(
"$info

Usage: perl $script_name [OPTIONS]

General options:
  --milter-name        name of mail filter in sendmail.cf
  --sendmail-config    path to the sendmail.cf

Additional options:
  --syslog-facility    syslog facility
  --verbose
  --use-verify
  --max-attempts=<count>
  --rcpt-tempfail=accept|tempfail

Miscellaneous:
  --version            display the version of $script_name and exit
  --help               display this help and exit
");
        exit();
    }
    if ($version) {
        print("$info\n");
        exit();
    }

    $syslog_param{'facility'} = $default_syslog_facility unless ($syslog_param{'facility'});
    $max_attempts = $default_max_attempts unless ($max_attempts);
    if ($rcpt_tempfail_accept =~ /^accept$/) {
	$rcpt_tempfail_accept = 1;
    } elsif ($rcpt_tempfail_accept =~ /^accept$/) {
	$rcpt_tempfail_accept = 0;
    } else {
	$rcpt_tempfail_accept = $default_rcpt_tempfail_accept;
    }

    #    sendmail',        
    $sendmail_config = find_sendmail_config(@sendmail_config_defaults) unless ($sendmail_config);

    #     ,       
    unless ($milter_name and $sendmail_config) {
	print("name of mail filter in sendmail.cf is missing") unless ($milter_name);
        print("path to the sendmail.cf is missing") unless ($sendmail_config);
        print("

Usage: perl $script_name [OPTIONS]

Try `$script_name --help' for more options.

");
        exit();
    }

    setlogsock($syslog_param{'socket'});
    openlog($syslog_param{'ident'}, $syslog_param{'options'}, $syslog_param{'facility'});

    my($conn) = Sendmail::Milter::auto_getconn($milter_name, $sendmail_config);

    print_msg("Found connection info for '$milter_name': $conn");

    if ($conn =~ /^local:(.+)$/) {
	my $unix_socket = $1;
	if (-e $unix_socket) {
	    print_msg("Attempting to unlink UNIX socket '$conn'...");
	    if (unlink($unix_socket) == 0) {
		print_msg("failed");
		exit();
	    }
	    print_msg("successful");
	}
    }

    if (not Sendmail::Milter::auto_setconn($milter_name, $sendmail_config)) {
	print_msg("Failed to detect connection information");
    } elsif (not Sendmail::Milter::register($milter_name, \%my_callbacks, SMFI_CURR_ACTS)) {
	print_msg("Failed to register callbacks for '$milter_name'");
    } else {
	print_msg("Starting Sendmail::Milter $Sendmail::Milter::VERSION engine");
	if (Sendmail::Milter::main()) {
	    print_msg("Successful exit from the Sendmail::Milter engine");
	} else {
	    print_msg("Unsuccessful exit from the Sendmail::Milter engine");
	}
    }
    closelog();
}

sub print_msg() {
    my($msg) = @_;
    print($msg."\n");
    syslog('notice', $msg);
}

sub find_sendmail_config {
    my(@sendmail_config_defaults) = @_;
    my($file);
    foreach $file (@sendmail_config_defaults) {
        if (-f $file) {
            return($file);
        }
    }
    return('');
}
